#include <Arduino_GFX_Library.h>
#define TFT_BL 9
Arduino_DataBus *bus = new Arduino_ESP32SPI(3 /* DC */, 2 /* CS */, 1 /* SCK */, 21 /* MOSI */, -1 /* MISO */);
Arduino_GFX *gfx = new Arduino_GC9A01(bus, 4 /* RST */, 3 /* rotation */, true /* IPS */);
#include <TinyGPSPlus.h>
////////
#include <Arduino.h>
#include <Wire.h>
#include <WiFi.h>
#include <WebServer.h>
#include <LSM303AGR_ACC_Sensor.h>
#include <LSM303AGR_MAG_Sensor.h>

#if defined(ARDUINO_SAM_DUE)
#define DEV_I2C Wire1   //Define which I2C bus is used. Wire1 for the Arduino Due
#define SerialPort Serial
#else
#define DEV_I2C Wire    //Or Wire
#define SerialPort Serial
#endif

TinyGPSPlus gps;
HardwareSerial GPSSerial(1);

// Components.
LSM303AGR_ACC_Sensor Acc(&DEV_I2C);
LSM303AGR_MAG_Sensor Mag(&DEV_I2C);

WebServer server;

float magnetometerX, magnetometerY, magnetometerZ;
float accelerometerX, accelerometerY, accelerometerZ;
float temperature;
int spo2 = 0;
int heartRate = 0;
String motion = "No Motion Detected";
String location = "https://www.google.com/maps?q=";
float heading = 0;
String direction = "";
String latitude = "";
String longitude = "";

const char *ssid = "JioFiber-4G";
const char *password = "1234567890";
//******************************************************/
#define BACKGROUND BLACK
#define MARK_COLOR WHITE
#define SUBMARK_COLOR DARKGREY // LIGHTGREY
#define HOUR_COLOR WHITE
#define MINUTE_COLOR BLUE // LIGHTGREY
#define SECOND_COLOR RED

#define SIXTIETH 0.016666667
#define TWELFTH 0.08333333
#define SIXTIETH_RADIAN 0.10471976
#define TWELFTH_RADIAN 0.52359878
#define RIGHT_ANGLE_RADIAN 1.5707963






static uint8_t conv2d(const char *p)
{
  uint8_t v = 0;
  return (10 * (*p - '0')) + (*++p - '0');
}

static int16_t w, h, center;
static int16_t hHandLen, mHandLen, sHandLen, markLen;
static float sdeg, mdeg, hdeg;
static int16_t osx = 0, osy = 0, omx = 0, omy = 0, ohx = 0, ohy = 0; // Saved H, M, S x & y coords
static int16_t nsx, nsy, nmx, nmy, nhx, nhy;                         // H, M, S x & y coords
static int16_t xMin, yMin, xMax, yMax;                               // redraw range
static int16_t hh, mm, ss;
static unsigned long targetTime; // next action time

static int16_t *cached_points;
static uint16_t cached_points_idx = 0;
static int16_t *last_cached_point;




String getCardinalDirection(float heading) {
  if (heading >= 337.5 || heading < 22.5)
    return "N";
  else if (heading >= 22.5 && heading < 67.5)
    return "NE";
  else if (heading >= 67.5 && heading < 112.5)
    return "E";
  else if (heading >= 112.5 && heading < 157.5)
    return "SE";
  else if (heading >= 157.5 && heading < 202.5)
    return "S";
  else if (heading >= 202.5 && heading < 247.5)
    return "SW";
  else if (heading >= 247.5 && heading < 292.5)
    return "W";
  else if (heading >= 292.5 && heading < 337.5)
    return "NW";
}

void updateSensorData() {
  int32_t magnetometer[3];
  Mag.GetAxes(magnetometer);

  magnetometerX = magnetometer[0];
  magnetometerY = magnetometer[1];
  magnetometerZ = magnetometer[2];

  // Calculate heading using magnetometer
  heading = atan2(magnetometer[1], magnetometer[0]) * 180 / PI;
  if (heading < 0) heading += 360;

  // Determine direction
  direction = getCardinalDirection(heading);

  // Read accelerometer data
  int32_t accelerometer[3];
  Acc.GetAxes(accelerometer);

  // Calculate movement based on difference in accelerometer readings
  int32_t movement = abs(accelerometer[0] - accelerometerX) +
                     abs(accelerometer[1] - accelerometerY) +
                     abs(accelerometer[2] - accelerometerZ);
  
  // Update current accelerometer values
  accelerometerX = accelerometer[0];
  accelerometerY = accelerometer[1];
  accelerometerZ = accelerometer[2];

  // If movement is detected (threshold may need adjustment)
  if (movement > 100) {
    motion = "Motion Detected";
  } else {
    motion = "No Motion Detected";
  }

  // Read temperature data
  Acc.GetTemperature(&temperature);

  // Simulated values for SPO2 and heart rate
  spo2 = random(90, 101);
  heartRate = random(60, 101);
}




void handleRoot() {
  updateSensorData();  // Update sensor data
  updateGPSData();

  String headingContent = "Heading: " + String(heading) + " degrees";
  String directionContent = "Direction: " + direction;

  String content = R"(
<!DOCTYPE html>
<html>
<head>
  <meta charset="UTF-8" name="viewport" content="width=device-width, initial-scale=1">
  <style>
    body {
      font-family: Arial, sans-serif;
      margin: 20px;
    }
    .sensor-box {
      border: 2px solid #ddd;
      padding: 20px;
      margin: 10px;
      text-align: center;
      font-size: 24px;
      background-color: #f8f8f8;  /* Light cream color */
    }
    .magnetometer {
      color: #3366cc;  /* Dark blue color */
    }
    .accelerometer {
      color: #cc3366;  /* Dark pink color */
    }
    .temperature {
      color: #33cc33;  /* Dark green color */
    }
    .spo2 {
      color: #ff6600; /* Orange color */
    }
    .heart-rate {
      color: #cc0000; /* Red color */
    }
    .motion {
      color: #990099; /* Purple color */
    }
    .heading {
      color: #666666; /* Dark gray color */
    }
    .direction {
      color: #0066ff; /* Blue color */
    }
    .location {
      color: #0066ff; /* Blue color */
      text-decoration: underline;
      cursor: pointer;
    }
  </style>
  <script>
    function updateValues(data) {
      const [magnetometerX, magnetometerY, magnetometerZ, accelerometerX, accelerometerY, accelerometerZ, temperature, spo2, heartRate, motion, location, heading, direction] = data.split(",");
      document.getElementById("magnetometerX").innerText = magnetometerX;
      document.getElementById("magnetometerY").innerText = magnetometerY;
      document.getElementById("magnetometerZ").innerText = magnetometerZ;
      document.getElementById("accelerometerX").innerText = accelerometerX;
      document.getElementById("accelerometerY").innerText = accelerometerY;
      document.getElementById("accelerometerZ").innerText = accelerometerZ;
      document.getElementById("temperature").innerText = temperature;
      document.getElementById("spo2").innerText = spo2;
      document.getElementById("heartRate").innerText = heartRate;
      document.getElementById("motion").innerText = motion;
      document.getElementById("location").innerHTML = '<a href="' + location + '" target="_blank">' + location + '</a>';
      document.getElementById("heading").innerText = heading;
      document.getElementById("direction").innerText = direction;
    }

    function fetchData() {
      fetch("/sensor-values")
        .then(response => response.text())
        .then(data => updateValues(data));
    }

    setInterval(fetchData, 1000); // Update every 1000 milliseconds (1 second)
  </script>
</head>
<body>
<div class="sensor-box magnetometer">
  <strong>Magnetometer</strong><br>
  X: <span id="magnetometerX">0.0</span><br>
  Y: <span id="magnetometerY">0.0</span><br>
  Z: <span id="magnetometerZ">0.0</span><br>
</div>
<div class="sensor-box accelerometer">
  <strong>Accelerometer</strong><br>
  X: <span id="accelerometerX">0.0</span><br>
  Y: <span id="accelerometerY">0.0</span><br>
  Z: <span id="accelerometerZ">0.0</span><br>
</div>
<div class="sensor-box temperature">
  <strong>Temperature</strong><br>
  <span id="temperature">0.0</span> &#8451;<br>
</div>
<div class="sensor-box spo2">
  <strong>SPO2</strong><br>
  <span id="spo2">0</span> %<br>
</div>
<div class="sensor-box heart-rate">
  <strong>Heart Rate</strong><br>
  <span id="heartRate">0</span> BPM<br>
</div>
<div class="sensor-box motion">
  <strong>Motion</strong><br>
  <span id="motion">No Motion Detected</span><br>
</div>
<div class="sensor-box heading">
  <strong>Heading</strong><br>
  <span id="heading">0.0</span> degrees<br>
</div>
<div class="sensor-box direction">
  <strong>Direction</strong><br>
  <span id="direction">N</span><br>
</div>
<div class="sensor-box location">
  <strong>Location</strong><br>
  <span id="location" class="location">Click to view</span><br>
</div>
</body>
</html>
    )";
  server.send(200, "text/html", content);
}



void setup(void)
{
  Serial.begin(115200);
  GPSSerial.begin(9600, SERIAL_8N1, 44, 43); // Start GPS communication

  // Serial.setDebugOutput(true);
  // while(!Serial);
  Serial.println("Arduino_GFX Clock example");

#ifdef GFX_EXTRA_PRE_INIT
  GFX_EXTRA_PRE_INIT();
#endif

  // Init Display
  if (!gfx->begin())
  {
    Serial.println("gfx->begin() failed!");
  }
  gfx->fillScreen(BACKGROUND);

#ifdef GFX_BL
  pinMode(GFX_BL, OUTPUT);
  digitalWrite(GFX_BL, HIGH);
#endif

  // init LCD constant
  w = gfx->width();
  h = gfx->height();
  if (w < h)
  {
    center = w / 2;
  }
  else
  {
    center = h / 2;
  }
  hHandLen = center * 3 / 8;
  mHandLen = center * 2 / 3;
  sHandLen = center * 5 / 6;
  markLen = sHandLen / 6;
  cached_points = (int16_t *)malloc((hHandLen + 1 + mHandLen + 1 + sHandLen + 1) * 2 * 2);

  // Draw 60 clock marks
  draw_round_clock_mark(
      // draw_square_clock_mark(
      center - markLen, center,
      center - (markLen * 2 / 3), center,
      center - (markLen / 2), center);

  hh = conv2d(__TIME__);
  mm = conv2d(__TIME__ + 3);
  ss = conv2d(__TIME__ + 6);

  targetTime = ((millis() / 1000) + 1) * 1000;


  delay(1000);
  Serial.begin(115200);
  Serial.println();

  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(250);
    Serial.print(".");
  }
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  // Initialize I2C bus.
  DEV_I2C.begin();

  // Initialize components.
  Acc.begin();
  Acc.Enable();
  Acc.EnableTemperatureSensor();
  Mag.begin();
  Mag.Enable();

  // Handle root URL
  server.on("/", HTTP_GET, handleRoot);

  server.on("/sensor-values", HTTP_GET, []() {
    updateSensorData();
    String sensorValues = String(magnetometerX) + "," + String(magnetometerY) + "," + String(magnetometerZ) + "," +
                          String(accelerometerX) + "," + String(accelerometerY) + "," + String(accelerometerZ) + "," +
                          String(temperature) + "," + String(spo2) + "," + String(heartRate) + "," +
                          motion + "," + location + "," + String(heading) + "," + direction;
    server.send(200, "text/plain", sensorValues);
  });

  server.begin();
  
}

void loop()
{
    server.handleClient();

  unsigned long cur_millis = millis();
  if (cur_millis >= targetTime)
  {
    targetTime += 1000;
    ss++; // Advance second
    if (ss == 60)
    {
      ss = 0;
      mm++; // Advance minute
      if (mm > 59)
      {
        mm = 0;
        hh++; // Advance hour
        if (hh > 23)
        {
          hh = 0;
        }
      }
    }
  }

  // Pre-compute hand degrees, x & y coords for a fast screen update
  sdeg = SIXTIETH_RADIAN * ((0.001 * (cur_millis % 1000)) + ss); // 0-59 (includes millis)
  nsx = cos(sdeg - RIGHT_ANGLE_RADIAN) * sHandLen + center;
  nsy = sin(sdeg - RIGHT_ANGLE_RADIAN) * sHandLen + center;
  if ((nsx != osx) || (nsy != osy))
  {
    mdeg = (SIXTIETH * sdeg) + (SIXTIETH_RADIAN * mm); // 0-59 (includes seconds)
    hdeg = (TWELFTH * mdeg) + (TWELFTH_RADIAN * hh);   // 0-11 (includes minutes)
    mdeg -= RIGHT_ANGLE_RADIAN;
    hdeg -= RIGHT_ANGLE_RADIAN;
    nmx = cos(mdeg) * mHandLen + center;
    nmy = sin(mdeg) * mHandLen + center;
    nhx = cos(hdeg) * hHandLen + center;
    nhy = sin(hdeg) * hHandLen + center;

    // redraw hands
    redraw_hands_cached_draw_and_erase();

    ohx = nhx;
    ohy = nhy;
    omx = nmx;
    omy = nmy;
    osx = nsx;
    osy = nsy;

    delay(1);
  }

  
}

void draw_round_clock_mark(int16_t innerR1, int16_t outerR1, int16_t innerR2, int16_t outerR2, int16_t innerR3, int16_t outerR3)
{
  float x, y;
  int16_t x0, x1, y0, y1, innerR, outerR;
  uint16_t c;

  for (uint8_t i = 0; i < 60; i++)
  {
    if ((i % 15) == 0)
    {
      innerR = innerR1;
      outerR = outerR1;
      c = MARK_COLOR;
    }
    else if ((i % 5) == 0)
    {
      innerR = innerR2;
      outerR = outerR2;
      c = MARK_COLOR;
    }
    else
    {
      innerR = innerR3;
      outerR = outerR3;
      c = SUBMARK_COLOR;
    }

    mdeg = (SIXTIETH_RADIAN * i) - RIGHT_ANGLE_RADIAN;
    x = cos(mdeg);
    y = sin(mdeg);
    x0 = x * outerR + center;
    y0 = y * outerR + center;
    x1 = x * innerR + center;
    y1 = y * innerR + center;

    gfx->drawLine(x0, y0, x1, y1, c);
  }
}

void draw_square_clock_mark(int16_t innerR1, int16_t outerR1, int16_t innerR2, int16_t outerR2, int16_t innerR3, int16_t outerR3)
{
  float x, y;
  int16_t x0, x1, y0, y1, innerR, outerR;
  uint16_t c;

  for (uint8_t i = 0; i < 60; i++)
  {
    if ((i % 15) == 0)
    {
      innerR = innerR1;
      outerR = outerR1;
      c = MARK_COLOR;
    }
    else if ((i % 5) == 0)
    {
      innerR = innerR2;
      outerR = outerR2;
      c = MARK_COLOR;
    }
    else
    {
      innerR = innerR3;
      outerR = outerR3;
      c = SUBMARK_COLOR;
    }

    if ((i >= 53) || (i < 8))
    {
      x = tan(SIXTIETH_RADIAN * i);
      x0 = center + (x * outerR);
      y0 = center + (1 - outerR);
      x1 = center + (x * innerR);
      y1 = center + (1 - innerR);
    }
    else if (i < 23)
    {
      y = tan((SIXTIETH_RADIAN * i) - RIGHT_ANGLE_RADIAN);
      x0 = center + (outerR);
      y0 = center + (y * outerR);
      x1 = center + (innerR);
      y1 = center + (y * innerR);
    }
    else if (i < 38)
    {
      x = tan(SIXTIETH_RADIAN * i);
      x0 = center - (x * outerR);
      y0 = center + (outerR);
      x1 = center - (x * innerR);
      y1 = center + (innerR);
    }
    else if (i < 53)
    {
      y = tan((SIXTIETH_RADIAN * i) - RIGHT_ANGLE_RADIAN);
      x0 = center + (1 - outerR);
      y0 = center - (y * outerR);
      x1 = center + (1 - innerR);
      y1 = center - (y * innerR);
    }
    gfx->drawLine(x0, y0, x1, y1, c);
  }
}

void redraw_hands_cached_draw_and_erase()
{
  gfx->startWrite();
  draw_and_erase_cached_line(center, center, nsx, nsy, SECOND_COLOR, cached_points, sHandLen + 1, false, false);
  draw_and_erase_cached_line(center, center, nhx, nhy, HOUR_COLOR, cached_points + ((sHandLen + 1) * 2), hHandLen + 1, true, false);
  draw_and_erase_cached_line(center, center, nmx, nmy, MINUTE_COLOR, cached_points + ((sHandLen + 1 + hHandLen + 1) * 2), mHandLen + 1, true, true);
  gfx->endWrite();
}

void draw_and_erase_cached_line(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t color, int16_t *cache, int16_t cache_len, bool cross_check_second, bool cross_check_hour)
{
#if defined(ESP8266)
  yield();
#endif
  bool steep = _diff(y1, y0) > _diff(x1, x0);
  if (steep)
  {
    _swap_int16_t(x0, y0);
    _swap_int16_t(x1, y1);
  }

  int16_t dx, dy;
  dx = _diff(x1, x0);
  dy = _diff(y1, y0);

  int16_t err = dx / 2;
  int8_t xstep = (x0 < x1) ? 1 : -1;
  int8_t ystep = (y0 < y1) ? 1 : -1;
  x1 += xstep;
  int16_t x, y, ox, oy;
  for (uint16_t i = 0; i <= dx; i++)
  {
    if (steep)
    {
      x = y0;
      y = x0;
    }
    else
    {
      x = x0;
      y = y0;
    }
    ox = *(cache + (i * 2));
    oy = *(cache + (i * 2) + 1);
    if ((x == ox) && (y == oy))
    {
      if (cross_check_second || cross_check_hour)
      {
        write_cache_pixel(x, y, color, cross_check_second, cross_check_hour);
      }
    }
    else
    {
      write_cache_pixel(x, y, color, cross_check_second, cross_check_hour);
      if ((ox > 0) || (oy > 0))
      {
        write_cache_pixel(ox, oy, BACKGROUND, cross_check_second, cross_check_hour);
      }
      *(cache + (i * 2)) = x;
      *(cache + (i * 2) + 1) = y;
    }
    if (err < dy)
    {
      y0 += ystep;
      err += dx;
    }
    err -= dy;
    x0 += xstep;
  }
  for (uint16_t i = dx + 1; i < cache_len; i++)
  {
    ox = *(cache + (i * 2));
    oy = *(cache + (i * 2) + 1);
    if ((ox > 0) || (oy > 0))
    {
      write_cache_pixel(ox, oy, BACKGROUND, cross_check_second, cross_check_hour);
    }
    *(cache + (i * 2)) = 0;
    *(cache + (i * 2) + 1) = 0;
  }
}

void write_cache_pixel(int16_t x, int16_t y, int16_t color, bool cross_check_second, bool cross_check_hour)
{
  int16_t *cache = cached_points;
  if (cross_check_second)
  {
    for (uint16_t i = 0; i <= sHandLen; i++)
    {
      if ((x == *(cache++)) && (y == *(cache)))
      {
        return;
      }
      cache++;
    }
  }
  if (cross_check_hour)
  {
    cache = cached_points + ((sHandLen + 1) * 2);
    for (uint16_t i = 0; i <= hHandLen; i++)
    {
      if ((x == *(cache++)) && (y == *(cache)))
      {
        return;
      }
      cache++;
    }
  }
  gfx->writePixel(x, y, color);
}

void updateGPSData()
{
   while (GPSSerial.available() > 0)
  {
    char c = GPSSerial.read();
    gps.encode(c);
  }

  if (gps.location.isValid())
  {
    latitude = String(gps.location.lat(), 6);
    longitude = String(gps.location.lng(), 6);
    location = "https://www.google.com/maps?q=" + latitude + "," + longitude;
  }
  else
  {
    latitude = "N/A";
    longitude = "N/A";
    location = "Location not found";
  }
}
